from django.contrib.auth import get_user_model
from django.core.validators import validate_email
from django.contrib.auth.backends import ModelBackend, BaseBackend
from django.core.exceptions import ValidationError, PermissionDenied
from django.utils.translation import gettext as _
from rest_framework.authentication import TokenAuthentication, exceptions
from ocs_authentication.util import generate_tokens, get_profile, create_or_update_user
from ocs_authentication.auth_profile.models import AuthProfile
from ocs_authentication.exceptions import ProfileException, OAuthTokenException


class OCSTokenAuthentication(TokenAuthentication):
    """
    This Allows authentication based on the api_key stored in the AuthProfile model.
    This should allow users to use the same api_key between client apps and the Oauth Server.
    TODO:: Once we switch to just using the DRF tokens rather than allowing both DRF tokens and
           the AuthProfile api_tokens, this backend should no longer be necessary.
    """
    def authenticate_credentials(self, key):
        try:
            output = super().authenticate_credentials(key)
            return output
        except exceptions.AuthenticationFailed:
            pass
        # Fallback on trying the api_token in the AuthToken model
        try:
            token = AuthProfile.objects.select_related('user').get(api_token=key)
        except AuthProfile.DoesNotExist:
            raise exceptions.AuthenticationFailed(_('Invalid token.'))

        if not token.user.is_active:
            raise exceptions.AuthenticationFailed(_('User inactive or deleted.'))

        return (token.user, token)


class OAuthUsernamePasswordBackend(ModelBackend):
    """
    Authenticate against the OAuth Authorization server using
    grant_type: password

    This backend should be placed after a backend that checks the local database for if the user exists there.
    """
    def authenticate(self, request, username=None, password=None):
        try:
            access_token, refresh_token = generate_tokens(username, password)
        except OAuthTokenException:
            # The authorization server failed to generate tokens. The username and password still might be
            # able to authenticate via another backend, so return `None`.
            return None

        try:
            profile = get_profile(access_token=access_token)
        except ProfileException:
            # Failed to get profile data using newly created access token. Something is wrong, indicate not authorized.
            raise PermissionDenied('Failed to access user profile')

        return create_or_update_user(profile, password)

    def get_user(self, user_id):
        try:
            return get_user_model().objects.get(pk=user_id)
        except get_user_model().DoesNotExist:
            return None


class EmailOrUsernameModelBackend(BaseBackend):
    """
    Authenticate either with username and password, or with email and password.
    """
    def authenticate(self, request, username=None, password=None):
        is_email = True
        try:
            validate_email(username)
        except ValidationError:
            is_email = False
        if is_email:
            kwargs = {'email': username}
        else:
            kwargs = {'username': username}
        try:
            user = get_user_model().objects.get(**kwargs)
            if user.check_password(password):
                return user
        except get_user_model().DoesNotExist:
            return None

    @staticmethod
    def get_user(user_id):
        try:
            return get_user_model().objects.get(pk=user_id)
        except get_user_model().DoesNotExist:
            return None



import json
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status

from ocs_authentication.permissions import IsServer
from ocs_authentication.util import create_or_update_user, Profile


class AddUpdateUserView(APIView):
    """
    This view is meant to be called by the Oauth Server when a new user account is created. This will create
    the corresponding user account within this Oauth client app and give it the same api-token, so the user
    can access this application with their api-token without needing to initially login with their password.
    This should also be called on token change or on any user info change.
    """
    permission_classes = [IsServer]

    def post(self, request):
        data = json.loads(request.body.decode('utf-8'))
        profile = Profile(
            data.get('first_name', ''),
            data.get('last_name', ''),
            data.get('username', ''),
            data.get('email', ''),
            data.get('tokens', {}).get('api_token', ''),
            data.get('is_staff', False),
            data.get('is_superuser', False),
            data.get('profile', {}).get('staff_view', False)
        )
        # The password will not be set here since this only has the profile api info.
        # The password will only get set when logging in using username/password auth
        # which is forwarded through Oauth.
        create_or_update_user(profile, password=None)
        return Response({'message': 'User account updated'}, status=status.HTTP_200_OK)




from rest_framework import permissions
from django.conf import settings


class IsAdminOrReadOnly(permissions.BasePermission):
    """The request is either read-only, or the user is staff"""
    def has_permission(self, request, view):
        return bool(
            request.method in permissions.SAFE_METHODS
            or request.user and request.user.is_staff
        )


class IsServer(permissions.BasePermission):
    message = 'Invalid or missing API Key.'

    def has_permission(self, request, view):
        authorization = request.META.get("HTTP_AUTHORIZATION")

        key = ''
        if authorization:
            try:
                _, key = authorization.split("Server ")
            except ValueError:
                pass

        if key:
            return key == settings.OCS_AUTHENTICATION['OAUTH_SERVER_KEY']

        return False



from django.db import models
from django.conf import settings as django_settings

class AuthProfile(models.Model):
    user = models.OneToOneField(django_settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    staff_view = models.BooleanField(default=False)
    api_token = models.CharField(max_length=255, default='')



from setuptools import setup, find_packages

# Read the contents of the README
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ocs-authentication',
    use_scm_version=True,
    description='Authentication backends and utilities for the OCS applications',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/observatorycontrolsystem/ocs-authentication',
    packages=find_packages(),
    python_requires='>=3.7',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],
    setup_requires=['setuptools_scm'],
    install_requires=[
        'django>=3.2,<4.0',
        'djangorestframework>=3.12,<3.13',
        'requests>=2.22,<2.27',
    ],
    extras_require={
        'tests': ['pytest']
    }
)



"""
Django settings for ocs_authentication project.

Generated by 'django-admin startproject' using Django 3.2.8.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""

from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = 'abcdefg'

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework.authtoken',
    'ocs_authentication.auth_profile'
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}



from dataclasses import dataclass

import requests
from django.db import transaction
from django.contrib.auth import get_user_model

from ocs_authentication.settings import ocs_auth_settings
from ocs_authentication.auth_profile.models import AuthProfile
from ocs_authentication.exceptions import ProfileException, OAuthTokenException


@dataclass
class Profile:
    """Dataclass encapsulating profile information"""
    first_name: str
    last_name: str
    username: str
    email: str
    api_token: str
    is_staff: bool
    is_superuser: bool
    staff_view: bool


def get_profile(access_token: str) -> Profile:
    headers = {'Authorization': f'Bearer {access_token}'}
    profile_response = requests.get(
        ocs_auth_settings.OAUTH_PROFILE_URL,
        headers=headers,
        timeout=ocs_auth_settings.REQUESTS_TIMEOUT_SECONDS
    )
    if profile_response.status_code == 200:
        return Profile(
            profile_response.json()['first_name'],
            profile_response.json()['last_name'],
            profile_response.json()['username'],
            profile_response.json()['email'],
            profile_response.json()['tokens']['api_token'],
            profile_response.json()['is_staff'],
            profile_response.json()['is_superuser'],
            profile_response.json()['profile']['staff_view'],
        )
    else:
        raise ProfileException('Unable to access profile information')


def generate_tokens(username: str, password: str):
    token_response = requests.post(
        ocs_auth_settings.OAUTH_TOKEN_URL,
        data={
            'grant_type': 'password',
            'username': username,
            'password': password,
            'client_id': ocs_auth_settings.OAUTH_CLIENT_ID,
            'client_secret': ocs_auth_settings.OAUTH_CLIENT_SECRET
        },
        timeout=ocs_auth_settings.REQUESTS_TIMEOUT_SECONDS
    )
    if token_response.status_code == 200:
        return token_response.json()['access_token'], token_response.json()['refresh_token']
    else:
        raise OAuthTokenException('Failed to generate OAuth tokens')


def create_or_update_user(profile: Profile, password: str):
    with transaction.atomic():
        user, _ = get_user_model().objects.update_or_create(
            username=profile.username,
            defaults={
                'first_name': profile.first_name,
                'last_name': profile.last_name,
                'email': profile.email,
                'is_staff': profile.is_staff,
                'is_superuser': profile.is_superuser,
            }
        )
        if password:
            user.set_password(password)
        user.save()
        AuthProfile.objects.update_or_create(
            user=user,
            defaults={
                'staff_view': profile.staff_view,
                'api_token': profile.api_token
            }
        )
        # TODO:: This could will update DRFs internal authtoken as well. Uncomment this when we are
        #        ready to transition to a single api_token.
        # token, _ = Token.objects.get_or_create(user=user)
        # if profile.api_token != token.key:
        #     # Need to set the api token to some expected value
        #     token.delete()
        #     Token.objects.create(user=user, key=profile.api_token)
        return user
